/**
 * @file nvram.h
 * @brief nvram access utility for IBM ppc64 platforms.
 *
 * Copyright (c) 2003, 2004 International Business Machines
 * Common Public License Version 1.0 (see COPYRIGHT)
 *
 * @author Nathan Fontenot
 * @author Michael Strosaker
 * @author Todd Inglett
 */

#ifndef _DEV_NVRAM_H_
#define _DEV_NVRAM_H_

#define NVRAM_SIG_SP	0x02	/**< support processor signature */
#define NVRAM_SIG_OF	0x50	/**< open firmware config signature */
#define NVRAM_SIG_FW	0x51	/**< general firmware signature */
#define NVRAM_SIG_HW	0x52	/**< hardware (VPD) signature */
#define NVRAM_SIG_SYS	0x70	/**< system env vars signature */
#define NVRAM_SIG_CFG	0x71	/**< config data signature */
#define NVRAM_SIG_ELOG	0x72	/**< error log signature */
#define NVRAM_SIG_VEND	0x7e	/**< vendor defined signature */
#define NVRAM_SIG_FREE	0x7f	/**< Free space signature */
#define NVRAM_SIG_OS	0xa0	/**< OS defined signature */

/**
 * @def printmap(ch)
 * @brief dertermines if 'ch' is a printable character
 */
#define printmap(ch)	(isgraph(ch) ? (ch) : '.')

#define NVRAM_BLOCK_SIZE	16
#define NVRAM_FILENAME1		"/dev/nvram"
#define NVRAM_FILENAME2		"/dev/misc/nvram"

typedef unsigned short u16;
typedef unsigned int u32;

/**
 * @def MAX_CPUS
 * @brief maximum number of CPUS for errlog dumps
 */
#define MAX_CPUS 128

/**
 * @struct partition_header
 * @brief nvram partition header data
 */
struct partition_header {
    unsigned char	signature;  /**< partition signature */
    unsigned char 	checksum;   /**< partition checksum */
    unsigned short 	length;     /**< partition length */
    char 		name[12];
};

/* Internal representation of NVRAM. */
#define MAX_PARTITIONS 50
/**
 * @struct nvram
 * @brief internal representation of nvram data
 */
struct nvram {
    char 	*filename;		/**< original filename */
    int		fd;			/**< file descriptor */
    int 	nparts;			/**< number of partitions */
    int 	nbytes;			/**< size of data in bytes.  This 
                                         *   cannot be changed 
                                         *   (i.e. hardware size) 
					 */
    struct partition_header *parts[MAX_PARTITIONS]; 
                                        /**< partition header pointers 
                                         *   into data 
                                         */
    char	*data;                  /**< nvram contents */
};

/**
 * @struct rtas_error
 * @brief RTAS header definition
 */ 
struct rtas_error {
    int version:8;
    int severity:3;
    int rtas_disposition:2;
    int optional_part_presence:1;
    int reserved1:2;
    int initiator:4;
    int target:4;
    int errtype:8;
    int extended_error_length:32;
};

/**
 * @struct ext_hdr
 * @brief RTAS extended header definition
 */
struct ext_hdr {
    unsigned int valid:1;
    unsigned int unrecoverable:1;
    unsigned int recoverable:1;
    unsigned int unrecoverable_bypassed:1;  /**< i.e. degraded performance */
    unsigned int predictive:1;
    unsigned int newlog:1;
    unsigned int bigendian:1;	    /**< always 1 */
    unsigned int reserved1:1;

    unsigned int platform_specific:1;
    unsigned int reserved2:3;
    unsigned int platform_value:4;  /**< valid iff platform_specific */

    unsigned int power_pc:1;	    /**< always 1 */
    unsigned int reserved3:2;
    unsigned int addr_invalid:1;    /**< failing_address is invalid */
    unsigned int format_type:4;
#define EXTLOG_FMT_CPU 1
#define EXTLOG_FMT_MEMORY 2
#define EXTLOG_FMT_IO 3
#define EXTLOG_FMT_POST 4
#define EXTLOG_FMT_ENV 5
#define EXTLOG_FMT_POW 6
#define EXTLOG_FMT_IBMDIAG 12
#define EXTLOG_FMT_IBMSP 13

    unsigned int non_hardware:1;    /**< Firmware or software is suspect */
    unsigned int hot_plug:1;	    /**< Failing component may be hot plugged */
    unsigned int group_failure:1;   /**< Group of components should be replaced */
    unsigned int reserved4:1;
    unsigned int residual:1;	    /**< Residual error from previous boot 
				     *   (maybe a crash) 
				     */
    unsigned int boot:1;	    /**< Error during boot */
    unsigned int config_change:1;   /**< Configuration changed since last boot*/
    unsigned int post:1;	    /**< Error during POST */

    unsigned int bcdtime:32;	    /**< Time of error in BCD HHMMSS00 */
    unsigned int bcddate:32;	    /**< Time of error in BCD YYYYMMDD */
};

/**
 * @struct cpu_hdr
 * @brief RTAS CPU event
 */
struct cpu_hdr {
    unsigned int internal:1;
    unsigned int intcache:1;
    unsigned int extcache_parity:1; /**< or multi-bit ECC */
    unsigned int extcache_ecc:1;
    unsigned int sysbus_timeout:1;
    unsigned int io_timeout:1;
    unsigned int sysbus_parity:1;
    unsigned int sysbus_protocol:1;

    unsigned int cpuid:8;
    unsigned int element:16;
    unsigned int failing_address_hi:32;
    unsigned int failing_address_lo:32;
    /* 16 bytes reserved here */
};

/**
 * @struct mem_hdr
 * @brief RTAS memory event
 */
struct mem_hdr {
    unsigned int uncorrectable:1;
    unsigned int ECC:1;
    unsigned int threshold_exceeded:1;
    unsigned int control_internal:1;
    unsigned int bad_address:1;
    unsigned int bad_data:1;
    unsigned int bus:1;
    unsigned int timeout:1;
	
    unsigned int sysbus_parity:1;
    unsigned int sysbus_timeout:1;
    unsigned int sysbus_protocol:1;
    unsigned int hostbridge_timeout:1;
    unsigned int hostbridge_parity:1;
    unsigned int reserved1:1;
    unsigned int support:1;
    unsigned int sysbus_internal:1;

    unsigned int mem_controller_detected:8;	/**< who detected fault? */
    unsigned int mem_controller_faulted:8;	/**< who caused fault? */
	
    unsigned int failing_address_hi:32;
    unsigned int failing_address_lo:32;
	
    unsigned int ecc_syndrome:16;
    unsigned int memory_card:8;
    unsigned int reserved2:8;
    unsigned int sub_elements:32;		/**< one bit per element */
    unsigned int element:16;
};

/**
 * @struct io_hdr
 * @brief RTAS I/O event
 */
struct io_hdr {
    unsigned int bus_addr_parity:1;
    unsigned int bus_data_parity:1;
    unsigned int bus_timeout:1;
    unsigned int bridge_internal:1;
    unsigned int non_pci:1;		/**< i.e. secondary bus such as ISA */
    unsigned int mezzanine_addr_parity:1;
    unsigned int mezzanine_data_parity:1;
    unsigned int mezzanine_timeout:1;

    unsigned int bridge_via_sysbus:1;
    unsigned int bridge_via_mezzanine:1;
    unsigned int bridge_via_expbus:1;
    unsigned int detected_by_expbus:1;
    unsigned int expbus_data_parity:1;
    unsigned int expbus_timeout:1;
    unsigned int expbus_connection_failure:1;
    unsigned int expbus_not_operating:1;

    /* IOA signalling the error */
    unsigned int pci_sig_busno:8;
    unsigned int pci_sig_devfn:8;
    unsigned int pci_sig_deviceid:16;
    unsigned int pci_sig_vendorid:16;
    unsigned int pci_sig_revisionid:8;
    unsigned int pci_sig_slot:8;	/**< 00 => system board, ff => multiple */

    /* IOA sending at time of error */
    unsigned int pci_send_busno:8;
    unsigned int pci_send_devfn:8;
    unsigned int pci_send_deviceid:16;
    unsigned int pci_send_vendorid:16;
    unsigned int pci_send_revisionid:8;
    unsigned int pci_send_slot:8;	/**< 00 => system board, ff => multiple */
};
					
/**
 * @struct sp_hdr
 * @brief RTAS service processor log event
 */
struct sp_hdr {
    unsigned int ibm:32;	/**< "IBM\0" */

    unsigned int timeout:1;
    unsigned int i2c_bus:1;
    unsigned int i2c_secondary_bus:1;
    unsigned int sp_memory:1;
    unsigned int sp_registers:1;
    unsigned int sp_communication:1;
    unsigned int sp_firmware:1;
    unsigned int sp_hardware:1;

    unsigned int vpd_eeprom:1;
    unsigned int op_panel:1;
    unsigned int power_controller:1;
    unsigned int fan_sensor:1;
    unsigned int thermal_sensor:1;
    unsigned int voltage_sensor:1;
    unsigned int reserved1:2;

    unsigned int serial_port:1;
    unsigned int nvram:1;
    unsigned int rtc:1;
    unsigned int jtag:1;
    unsigned int tod_battery:1;
    unsigned int reserved2:1;
    unsigned int heartbeat:1;
    unsigned int surveillance:1;

    unsigned int pcn_connection:1;	/**< power control network */
    unsigned int pcn_node:1;
    unsigned int reserved3:2;
    unsigned int pcn_access:1;
    unsigned int reserved:3;

    unsigned int sensor_token:32;	/**< zero if undef */
    unsigned int sensor_index:32;	/**< zero if undef */
};

/**
 * @var severity_names
 * @brief conversion for RTAS severities to human readable strings
 */
char *severity_names[] = {
    "NO ERROR", "EVENT", "WARNING", "ERROR_SYNC", "ERROR", "FATAL", "(6)", "(7)"
};

/**
 * @var rtas_disposition_names
 * @brief conversion for RTAS dispositions to human readable strings
 */
char *rtas_disposition_names[] = {
    "FULLY RECOVERED", "LIMITED RECOVERY", "NOT RECOVERED", "(4)"
};

/**
 * @var entity_names
 * @brief conversion for RTAS initiators and targets to human readable strings
 */
char *entity_names[] = { /* for initiator & targets */
    "UNKNOWN", "CPU", "PCI", "ISA", "MEMORY", "POWER MANAGEMENT", "HOT PLUG", 
    "(7)", "(8)", "(9)", "(10)", "(11)", "(12)", "(13)", "(14)", "(15)" 
};

/**
 * @var descs
 * @brief Array of VPD field names and descriptions
 */
static struct { char *name; char *desc; } descs[] = {
    {"PN", "Part Number"},
    {"FN", "FRU Number"},
    {"EC", "EC Level"},
    {"MN", "Manufacture ID"},
    {"SN", "Serial Number"},
    {"LI", "Load ID"},
    {"RL", "ROM Level"},
    {"RM", "Alterable ROM Level"},
    {"NA", "Network Address"},
    {"DD", "Device Driver Level"},
    {"DG", "Diagnostic Level"},
    {"LL", "Loadable Microcode Level"},
    {"VI", "Vendor ID/Device ID"},
    {"FU", "Function Number"},
    {"SI", "Subsystem Vendor ID/Device ID"},
    {"VK", "Platform"},			/**< "RS6K" => VPD is present */
    {"TM", "Model"},			/**< IBM specific? */
    {"YL", "Location Code"},		/**< IBM specific? */
    {"BR", "Brand"},			/**< IBM specific */
    {"CI", "CEC ID"},			/**< IBM specific */
    {"RD", "Rack ID"},			/**< IBM specific */
    {"PA", "Op Panel Installed"},	/**< IBM specific */
    {"NN", "Node Name"},		/**< IBM specific */
};

#endif
